open Xlist
open Xstring
open Interfaces
open Definitions_abstract
open Definitions_lift_subst

open SizedList

class fv_permut_fun _gmod =
object (self)
	inherit lift_fun _gmod
	
	(** the definition title *)
	method title : string = "Free variable index permutations"
	(** function parameters list (name * type_name) *)
	method fun_params : (string * string) list =
		("_k", "nat") ::
		("_l_" ^ self#get_target#get_name, "list nat") ::
		("_m", "nat") :: []
	
	(** name of the function in a given target grammar *)
	method fun_name_type_name (name : string) : string =
		self#get_source#get_name ^ "_fv_permut_in_" ^ name
	
	(** generation of parameters *)
	method private parameter_function_gen (p : parameter) : string =
		let lists_param =
		("_l_" ^ self#get_target#get_name) :: [] in

		let shift =
			let raw_shifts = List.map self#shift_fun_gen p#get_shifts in
			let non_empty_shifts = List.filter (fun s -> s <> "") raw_shifts in
			String.concat " + " non_empty_shifts in
		let subst =
			("#param#", p#get_name) ::
			("#recfun#", self#fun_name_type_name p#get_type#get_name) ::
			("#shift#", if shift = "" then "_m" else "(_m + " ^ shift ^ ")") ::
			("#lists_param#", String.concat " " lists_param) ::
			[]
		and text =
			if self#is_self_rec p || self#is_back_rec p || self#is_throughs_rec p
			then
				"(#recfun# _k #lists_param# #shift# #param#)"
			else
				"#param#" in
		Xstring.replace subst text
	
	(** generation of constructors rhs : index case *)
	method private index_constructor_rhs_function_gen (c : constructor) : string list =
		let subst =
			("#target_type#", self#get_target#get_name) ::
			("#constr#", c#get_name) ::
			("#index#", c#get_index_name) ::
			[]
		and text =
			if self#get_target = self#get_source
			then
				"if le_lt_dec _m #index#" ::
				"then" ::
				"  if le_lt_dec (List.length _l_#target_type#) (#index# - _m)" ::
				"  then #constr# (#index# + _k)" ::
				"  else #constr# (_m + List.nth (#index# - _m) _l_#target_type# 0)" ::
				"else #constr# #index#" ::
				[]
			else "#constr# #index#" :: [] in
		Xstring.replaces subst [] text

	(** generation of index lemmas and tactics *)
	method index_lemmas_and_tactics (g : grammar) : (string list * string list) =
		let _ = self#set_target g in
		let subst =
			("#fv_permut#", self#fun_name) ::
			("#constr#",
				match g#get_index_constructor_name with
					Some s -> s
				| None -> failwith ("Definitions_others.fv_permut_fun.index_lemmas_and_tactics : no index constructor in " ^ g#get_name)) ::
			[]
		and text =
			"  Lemma #fv_permut#_#constr#_dec :" ::
			"    forall (_k : nat) (_l : list nat) (_m _p : nat)," ::
			"    { prod (prod (_m <= _p) ((_p - _m) < List.length _l)) (#fv_permut# _k _l _m (#constr# _p) = #constr# (_m + List.nth (_p - _m) _l 0)) } +" ::
			"    { prod (prod (_m <= _p) (List.length _l <= (_p - _m))) (#fv_permut# _k _l _m (#constr# _p) = #constr# (_p + _k)) } +" ::
			"    { prod (_p < _m) (#fv_permut# _k _l _m (#constr# _p) = #constr# _p) }." ::
			"  Proof." ::
			"    intros; simpl; destruct le_lt_dec; try destruct le_lt_dec; subst; simpl; auto." ::
			"  Qed." ::
			"" ::
			"  Ltac _#fv_permut#_#constr#_dec_tac k l m p :=" ::
			"    let Hcomp := fresh \"Hcomp\" in" ::
			"       let Hcomp' := fresh \"Hcomp'\" in" ::
			"          let Hfun := fresh \"Hfun\" in" ::
			"            destruct (#fv_permut#_#constr#_dec k l m p) as" ::
			"              [ [ [ [ Hcomp Hcomp' ] Hfun ] | [ [ Hcomp Hcomp' ] Hfun ] ] | [ Hcomp Hfun ] ];" ::
			"              rewrite ?Hfun in *; clear Hfun." ::
			"" ::
			[]
		and text2 =
			"      | [ |- context [ #fv_permut# ?k ?l ?m (#constr# ?p) ] ] =>" ::
			"        _#fv_permut#_#constr#_dec_tac k l m p" ::
			"" ::
			"      | [ H : context [ #fv_permut# ?k ?l ?m (#constr# ?p) ] |- _ ] =>" ::
			"        _#fv_permut#_#constr#_dec_tac k l m p" ::
			"" ::
			[] in
		(Xstring.replaces subst [] text, Xstring.replaces subst [] text2)
	
	(** generation of index equation lemmas *)
	method index_equation_lemmas (g : grammar) : (string list * string list) =
		let _ = self#set_target g in
		let subst =
			("#fv_permut#", self#fun_name) ::
			("#constr#",
				match g#get_index_constructor_name with
					Some s -> s
				| None -> failwith ("Definitions_others.fv_permut_fun.index_lemmas_and_tactics : no index constructor in " ^ g#get_name)) ::
			[]
		and text =
			"  Lemma #fv_permut#_#constr#_le_le :" ::
			"    forall (_k : nat) (_l : list nat) (_m _p : nat)," ::
			"      _m <= _p -> List.length _l <= (_p - _m) -> #fv_permut# _k _l _m (#constr# _p) = #constr# (_p + _k)." ::
			"  Proof." ::
			"    intros; simpl; destruct le_lt_dec; try destruct le_lt_dec; f_equal; omega." ::
			"  Qed." ::
			"" ::
			"  Lemma #fv_permut#_#constr#_le_lt :" ::
			"    forall (_k : nat) (_l : list nat) (_m _p : nat)," ::
			"      _m <= _p -> (_p - _m) < List.length _l -> #fv_permut# _k _l _m (#constr# _p) = #constr# (_m + List.nth (_p - _m) _l 0)." ::
			"  Proof." ::
			"    intros; simpl; destruct le_lt_dec; try destruct le_lt_dec; f_equal; omega." ::
			"  Qed." ::
			"" ::
			"  Lemma #fv_permut#_#constr#_lt :" ::
			"    forall (_k : nat) (_l : list nat) (_m _p : nat)," ::
			"      _p < _m -> #fv_permut# _k _l _m (#constr# _p) = #constr# _p." ::
			"  Proof." ::
			"    intros; simpl; destruct le_lt_dec; try destruct le_lt_dec; f_equal; omega." ::
			"  Qed." ::
			"" ::
			[]
		and text2 =
			"      | [ |- context [ #fv_permut# ?k ?l ?m (#constr# ?p) ] ] =>" ::
			"          (cutrewrite ->" ::
			"            (#fv_permut# k l m (#constr# p) = #constr# (p + k));" ::
			"            [ | apply (#fv_permut#_#constr#_le_le k l m p); cfail ]) ||" ::
			"          (cutrewrite ->" ::
			"            (#fv_permut# k l m (#constr# p) = #constr# (m + List.nth (p - m) l 0));" ::
			"            [ | apply (#fv_permut#_#constr#_le_lt k l m p); cfail ]) ||" ::
			"          (cutrewrite ->" ::
			"            (#fv_permut# k l m (#constr# p) = #constr# p);" ::
			"            [ | apply (#fv_permut#_#constr#_lt k l m p); cfail ]) ||" ::
			"          fv_permut_index_dec_tac" ::
			"" ::
			"      | [ H : context [ #fv_permut# ?k ?l ?m (#constr# ?p) ] |- _ ] =>" ::
			"          (cutrewrite ->" ::
			"            (#fv_permut# k l m (#constr# p) = #constr# (p + k)) in H;" ::
			"            [ | apply (#fv_permut#_#constr#_le_le k l m p); cfail ]) ||" ::
			"          (cutrewrite ->" ::
			"            (#fv_permut# k l m (#constr# p) = #constr# (m + List.nth (p - m) l 0)) in H;" ::
			"            [ | apply (#fv_permut#_#constr#_le_lt k l m p); cfail ]) ||" ::
			"          (cutrewrite ->" ::
			"            (#fv_permut# k l m (#constr# p) = #constr# p) in H;" ::
			"            [ | apply (#fv_permut#_#constr#_lt k l m p); cfail ]) ||" ::
			"          fv_permut_index_dec_tac" ::
			"" ::
			[] in
		(Xstring.replaces subst [] text, Xstring.replaces subst [] text2)
end

class fv_permut_fun_with_source_and_target _gmod _source _target =
object (self)
	inherit fv_permut_fun _gmod as super
	
	initializer
	begin
		super#set_source _source;
		super#set_target _target
	end
end

class named_terms _gmod =
object (self)
	inherit abstract_terms_definition _gmod as super
	
	(** the definition title *)
	method title : string = "Named syntax"
	(** the category of generated index case *)
	method index_type : string = "string"
	(** extra definition lines to be inserted before the grammar definitions *)
	method coq_predefs : string list =
		SizedList.sizedlist @
		"" ::
		"  Require Import String." ::
		"  Local Open Scope string_scope." ::
		"" ::
		"  Definition _name_list := SizedList.sized_list string." ::
		"  Definition _xcons := @SizedList.scons string." ::
		"  Definition _xnil := @SizedList.snil string." ::
		"  Definition _index_of := SizedList.index_of string_dec." ::
		"  Definition _concat := @SizedList.app string." ::
		"  Definition _build_permut := SizedList.build_permut string_dec." ::
		"" ::
		[]
	
	(** extra definition lines to be inserted after the constructor cases *)
	method aux_defs : string list =
		(* we get the indexed grammar list *)
		let indexed = gmod#get_indexed_grams_through self#get_target in
		
		(* for each indexed grammar, we build the parameter *)
		let param_gen (g : grammar) =
			"(_xn_" ^ g#get_name ^ " : nat) " ^
			"(_xl_" ^ g#get_name ^ " : _name_list _xn_" ^ g#get_name ^ ") " in
			
		(* applying param_gen to every indexed grammars *)
		let params = List.map param_gen indexed in
		
		(* concatenation of params *)
		let all_params = String.concat "" params in
		
		(* output *)
		("  | _db_#raw_type# " ^ all_params ^ "(_arg : #raw_type#)") ::
		[]
	
	(** category names decoration (to avoid conflict) *)
	method decorate_type_name : string -> string =
		fun x -> if gmod#is_gmodule_gram x then "_" ^ x else x
	
	(** constructors names decoration (to avoid conflict) *)
	method decorate_constr_name : string -> string =
		fun x -> "_" ^ x
	
	(** generation of shifts *)
	method private shift_terms_gen (gram, level) : string =
		(* we get the indexed grammar list for this target *)
		let indexed = gmod#get_indexed_grams_through self#get_target in
		
		(* we look for the number of type_name among indexed *)
		let nb = Xlist.index_of gram indexed in
		
		(* conversion to string *)
		let nbs = if nb = 0 then "" else string_of_int nb in
		
		(* output *)
		"(xl" ^ nbs ^ " : _name_list " ^ level ^ ")"
	
	(** generation of parameters *)
	method parameter_terms_gen (p : parameter) : string =
		super#parameter_terms_gen p
	
end

class named_to_db_fun _gmod =
object (self)
	inherit abstract_fun_definition _gmod

	(** the definition title *)
	method title : string = "Named to De Bruijn translation function"
	
	(** extra definition lines to be inserted before the grammar definitions *)
	method coq_predefs : string list =
		[]
	
	(** extra definition lines to be inserted after the constructor cases *)
	method aux_defs : string list =
		(* we need the fv_permut function for this definition : this generates
		   the function name to use for a given source grammar *)
		let make_fv_permut (source : grammar) (rec_arg : string) : string =
			let gname = source#get_name in
			
			let fv_permut =
				new fv_permut_fun_with_source_and_target
					gmod source self#get_target in
					
			fv_permut#fun_name ^ " (_n_" ^ gname ^ " - _xn_" ^ gname ^ ") "
			^ "(_build_permut _n_" ^ gname ^ " _l_" ^ gname
			^ " _xn_" ^ gname ^ " _xl_" ^ gname ^ ")" ^ rec_arg in
		
		(* we get the indexed grammar list for this target *)
		let indexed = gmod#get_indexed_grams_through self#get_target in
		
		(* folds all indexed grammars to be lifted *)
		let rec lift_all_indexed l =
			match l with
					[] -> "_arg"
				| gram :: l ->
						"(" ^ make_fv_permut gram (" 0 " ^ lift_all_indexed l) ^ ")" in
		
		(* apply the previous folding to indexed *)
		let rhs = lift_all_indexed indexed in

		(* for each indexed grammar, we build the parameter *)
		let param_gen (g : grammar) =
			"_xn_" ^ g#get_name ^ " _xl_" ^ g#get_name ^ " " in
			
		(* applying param_gen to every indexed grammars *)
		let params = List.map param_gen indexed in
		
		(* concatenation of params *)
		let all_params = String.concat "" params in
						
		(* output *)
		("      | _db_#raw_target_type# " ^ all_params ^ "_arg =>") ::
		("        " ^ rhs) ::
		[]
	
	(** category names decoration (to avoid conflict) *)
	method decorate_type_name (name : string) : string =
		"_" ^ name
	
	(** constructors names decoration (to avoid conflict) *)
	method decorate_constr_name (name : string) : string =
		"_" ^ name
	
	(** function parameters list (name * type_name) *)
	method fun_params : (string * string) list =
		(* we get the indexed grammar list for this target *)
		let indexed = gmod#get_indexed_grams_through self#get_target in

		(* for each indexed grams throughs, builds the function parameters *)
		List.concat
			(List.map
					(fun g ->
								("_n_" ^ g#get_name, "nat") ::
								("_l_" ^ g#get_name, "_name_list _n_" ^ g#get_name) ::
								[])
					indexed)
	
	(** return type of the function *)
	method return_type : string =
		self#get_target#get_name
	
	(** name of the function in a given target grammar *)
	method fun_name_type_name (name : string) : string =
		"named_to_db_" ^ name
	
	(** the source grammars for which generetion is needed.
		* In this case : every indexed grammar. *)
	method get_source_grams : grammar option list =
		None :: []
	
	(** the target (processed) grammars for which generetion is needed.
		* By default, every grammar with the source grammar in its throughs. *)
	method get_target_grams (grams : grammar list) : grammar list =
		grams
	
	(** Unused virtual function *)
	method shift_fun_gen (type_name, level) : string = ""
	
	(** Generation of rhs for a given parameter (and auxiliary counter) *)
	method private param_rhs_function_gen (p : parameter) (n : int) : string * int =
		(* Generation function for a shift (with given counter for xlN) *)
		let shift_rhs_function_gen (type_name, level) nb : grammar * (string * string) =
			let nbs = if nb = 0 then "" else string_of_int nb in
			(type_name, (level, "xl" ^ nbs)) in
		
		(* we get the result of mapping shift_rhs_function_gen to all shifts *)
		let shifts =
			Xlist.map_count_start shift_rhs_function_gen n p#get_shifts in

		(* we get the indexed grammar list for this parameter *)
		let indexed = gmod#get_indexed_grams_through p#get_type in
		
		(* function that builds the inductive parameter for a given indexed
			grammar given the shifts for this parameter *)
		let make_args (gram : grammar) : string =
			(* builds the length additional argument " + ?" *)
			let arg_n =
				try
					let (_, (level, _)) = List.find (fun (g, _) -> g = gram) shifts in
					level					
				with
					(* if there is no shift for this grammar *)
					Not_found -> "" in
		
			(* builds the list argument with potential concatenation *)
			let arg_l =
				try
					let (_, (level, id)) = List.find (fun (g, _) -> g = gram) shifts in
					"(_concat " ^ level ^ " " ^ id ^ " " ^ "_n_" ^ gram#get_name ^ " _l_" ^ gram#get_name ^ ")"
				with
					(* if there is no shift for this grammar *)
					Not_found -> "_l_" ^ gram#get_name in

			let new_n =
				if arg_n = "" then "_n_" ^ gram#get_name
				else "(" ^ arg_n ^ " + _n_" ^ gram#get_name ^ ")" in
			new_n ^ " " ^ arg_l in

		let ordered_shifts =
			List.map make_args indexed in
		(String.concat " " ordered_shifts, List.length shifts)
	
	(** generation of parameters *)
	method parameter_function_gen (p : parameter) : string = ""
	method private parameter_function_gen_bis (p : parameter) (n : int) : string * int =
		let result_param = self#param_rhs_function_gen p n in
		let subst =
			("#param#", p#get_name) ::
			("#recfun#", self#fun_name_type_name p#get_type#get_name) ::
			("#list_param#", fst result_param) ::
			[]
		and text =
			if self#is_module_rec p
			then
				"(#recfun# #list_param# #param#)"
			else
				"#param#" in
		(Xstring.replace subst text, snd result_param)
	
	(** generation of constructors rhs : non index case *)
	method constructor_rhs_function_gen (c : constructor) : string list =
		[ String.concat " "
			(c#get_name :: (Xlist.map_count_var (fun p -> self#parameter_function_gen_bis p) c#get_params)) ]
	
	(** generation of constructors rhs : index case *)
	method index_constructor_rhs_function_gen (c : constructor) : string list =
		let subst =
			("#constr#", c#get_name) ::
			("#index#", c#get_index_name) ::
			[]
		and text =
			"#constr# (_index_of #index# _n_#raw_target_type# _l_#raw_target_type#)" ::
			[] in
		Xstring.replaces subst [] text
	
	method private param_lhs_function_gen (p : parameter) (nb : int) : string * int =
		let shift_lhs_function_gen (type_name, level) n : string =
			let nbs = if n = 0 then "" else string_of_int n in
			"xl" ^ nbs in
		let raw_shifts = Xlist.map_count_start shift_lhs_function_gen nb p#get_shifts in
		let non_empty_shifts = List.filter (fun s -> s <> "") raw_shifts in
		let shifts = String.concat " " non_empty_shifts in
		((if shifts = "" then p#get_name else shifts ^ " " ^ p#get_name),
		 List.length non_empty_shifts)
	
	(** generation of constructors *)
	method constructor_function_gen (c : constructor) : string list =
		let param_names =
			String.concat
				" " (Xlist.map_count_var self#param_lhs_function_gen c#get_params) in
		let subst =
			("#constr#", self#decorate_constr_name c#get_name) ::
			("#params#", param_names) ::
			[]
		and substs =
			(* let _ = self#init_nb in*)
			("#rhs#", List.map (fun s -> "        " ^ s) (self#rhs_function_gen c)) ::
			[]
		and text =
			"      | #constr# #params# =>" ::
			"#rhs#" ::
			[] in
		Xstring.replaces subst substs text
	
	method constructor_function_equation_gen (c : constructor) : string list = []
end

let aux_terms_definitions (gmod : gmodule) : terms_definition list =
	(new named_terms gmod :> terms_definition) ::
	[]

let aux_fun_definitions (gmod : gmodule) : fun_definition list =
	(new fv_permut_fun gmod :> fun_definition) ::
	(new named_to_db_fun gmod :> fun_definition) ::
	[]
