(* File lexer.mll *)
{
 open Parser
 open Str
 exception Eof
}
rule token = parse

| [' ' '\t' '\n']          { token lexbuf }

| '.'                      { DOT }
| ','                      { COMMA }
| ':'                      { COLON }
| '|'                      { BAR }
| '('                      { LPAR }
| ')'                      { RPAR }
| '['                      { LSQUARE }
| ']'                      { RSQUARE }

| '+'                      { PLUS }
| '-'                      { MINUS }
| '*'                      { MULT }

| "(* begin header *)" (_* as lxm) "(* end header *)"
                           { VERBATIM_HEADER (Str.split (regexp "\n") lxm) }
| "(* begin footer *)" (_* as lxm) "(* end footer *)"
                           { VERBATIM_FOOTER (Str.split (regexp "\n") lxm) }

| "(*"                     { LPARSTAR }
| "*)"                     { STARRPAR }
| ":="                     { ASSIGN }

| "Module"                 { MODULE }
| "Begin"                  { BEGIN }
| "End"                    { END }
| "verbatim"               { VERBATIM }
| "Inductive"              { INDUCTIVE }
| "Type"                   { TYPE }
| "with"                   { WITH }
| "index"                  { INDEX }
| "bind"                   { BIND }
| "in"                     { IN }

| ['0'-'9']+ as lxm        { INTEGER (int_of_string lxm) }
| ['A'-'z']['A'-'z' '0'-'9' '_']* as lxm
                           { ID (lxm) }

| eof                      { raise Eof }
