module Xio = struct
	
	(**
	* load a text file of a given name as a list of its lines.
	*
	* @params filename
	* @param filename the name of the file to load.
	*
	* @return the list of all the lines of filename.
	*)
	let load filename =
		(* open the file with given filename *)
		let channel = open_in filename in
		(* accumulator of lines *)
		let lines = ref [] in
		try
		(* accumulation loop, exits when exception End_of_file is raised *)
			while true; do
				lines := input_line channel :: !lines
			done; []
		with End_of_file ->
					begin
						(* close the file and revert the lines list *)
						close_in channel;
						List.rev !lines
					end
	;;
	
	(**
	* save a list of lines in a filename, replacing its previous content.
	*
	* @params filename lines
	* @param filename the name of the file that is to receive the text.
	* @param lines the text to save as a list of string.
	*)
	let save filename lines =
		let channel = open_out filename in
		let output_line s = output_string channel (s ^ "\n") in
		let _ = List.iter output_line lines in
		close_out channel
	
end
