Require Import Utf8.
Require Import Case_STLCletval.
Import STLCterms.

Local Open Scope string_scope.

Notation "'Λ' x '·' t" := (_lam (_xcons x 0 _xnil) t) (at level 50) : named_term_scope.
Notation "t '⊙' u" := (_app t u) (at level 55) : named_term_scope.
Notation "t '○' u" := (app t u) (at level 55) : named_term_scope.
Notation "'&' x" := (_var x) (at level 10) : named_term_scope.

Notation "t '[||]'" := (_db_term 0 _xnil t) (at level 10) : named_term_scope.
Notation "t '[|' x '|]'" := (_db_term 1 (_xcons x 0 _xnil) t) (at level 10) : named_term_scope.
Notation "'(!' t '!)'" := (named_to_db_term 0 _xnil t) (at level 0) : named_term_scope.

Bind Scope named_term_scope with _term.
Delimit Scope named_term_scope with _term.
Local Open Scope named_term_scope.

Example _id : _term := Λ"x"·&"x".
Example _apply : _term :=
  Λ"f"·Λ"x"·&"f" ⊙ &"x".
Example _delta : _term :=
  Λ"x"·&"x" ⊙ &"x".
Example _omega : _term := _delta ⊙ _delta.

Notation "t '[' u ']'" := (term_subst_in_term u 0 t) (at level 0).

Reserved Notation "t '|->' u" (at level 60).

Inductive red : term -> term -> Prop :=
| beta :
  forall (t u a : term),
    a = t[u] ->
    (lam t) ○ u |-> a

| eta :
  forall X (t a : term),
    a = (! Λ X·(t[||] ⊙ &X) !) ->
    a |-> t

| ctx_lam :
  forall t u: term,
    t |-> u ->
    lam t |-> lam u

| ctx_app_l :
  forall t' t u : term,
    t |-> u ->
      t ○ t' |-> u ○ t'

| ctx_app_r :
  forall t' t u : term,
    t |-> u ->
      t' ○ t |-> t' ○ u

        where "t '|->' u" := (red t u).

Reserved Notation "t '|-->' u" (at level 60).

Inductive reds : term -> term -> Prop :=
| reds_zero :
  forall t : term,
    t |--> t

| reds_step :
  forall t u v : term,
    t |-> u ->
      u |--> v ->
        t |--> v

          where "t '|-->' u" := (reds t u).

Definition _val (t : term) : _term :=
  Λ"_x"·(&"_x" ⊙ t[||]).

Definition val (t : term) : term :=
  (! _val t !).

Definition val' (t : term) : term :=
  lam (app (var 0) (term_lift_in_term 1 0 t)).

Lemma val_eq :
  forall t : term,
    val t = val' t.
Proof.
  intro t. unfold val, val'; simpl; crush_tac; dbgen_tac.
Qed.

Definition _letval Y (t : term) (u : _term) : _term :=
  Λ "__x"·(t[||] ⊙ Λ Y·(u ⊙ &"__x")).

Notation "'let_val' x '=_' t 'in_' u" := (_letval x t u) (at level 50) : named_term_scope.

Definition letval (t u : term) : term :=
  (! let_val "y" =_ t in_ (u[|"y"|]) !).

Definition letval' (t u : term) : term :=
  lam (app (term_lift_in_term 1 0 t)
    (lam (app (term_lift_in_term 1 1 u) (var 1)))).

Lemma letval_eq :
  forall t u : term,
    letval t u = letval' t u.
Proof.
  intros t u. unfold letval, letval', named_to_db_term.
  simpl; crush_tac. dbgen_tac.
Qed.

Lemma letval_val :
  forall (t u : term),
    (! let_val "y" =_ val t in_ (u[|"y"|]) !) |--> u[t].
Proof.
  intros t u. unfold _letval, val, _val. simpl.

  eapply reds_step. simpl; dbgen_tac.
  apply ctx_lam. apply beta. simpl; dbgen_tac.

  eapply reds_step. simpl; dbgen_tac.
  apply ctx_lam. apply beta. simpl; dbgen_tac.

  apply reds_step with (u[t]).
  apply (eta "__x"); simpl; dbgen_tac.

  apply reds_zero.
Qed.
